@@ -1,19 +1,21 @@
 import ast
 import copy
 import inspect
+import logging
 import sys
 import tempfile
 import textwrap
-import logging
 import warnings
-log = logging.getLogger(__name__)
 
 import astor
 from miniutils.opt_decorator import optional_argument_decorator
 from miniutils import magic_contract
 
-from pragma.core.stack import DictStack
-from pragma.core.resolve import resolve_literal, resolve_iterable, resolve_indexable, resolve_name_or_attribute, make_ast_from_literal
+from .stack import DictStack
+from .resolve import resolve_literal, resolve_iterable, resolve_indexable, resolve_name_or_attribute, \
+    make_ast_from_literal
+
+log = logging.getLogger(__name__)
 
 
 @magic_contract
@@ -32,7 +34,7 @@ def function_ast(f):
 
     try:
         found = inspect.findsource(f)
-    except IndexError as err:
+    except IndexError as err:  # pragma: nocover
         raise IOError((
             'Discrepancy in number of decorator @magics expected by '
             'inspect vs. __code__.co_firstlineno\n'
@@ -56,7 +58,8 @@ class DebugTransformerMixin:  # pragma: nocover
             if new_node is None:
                 log.debug("Deleted >>> {} <<<".format(orig_node_code))
             elif isinstance(new_node, ast.AST):
-                log.debug("Converted >>> {} <<< to >>> {} <<<".format(orig_node_code, astor.to_source(new_node).strip()))
+                log.debug(
+                    "Converted >>> {} <<< to >>> {} <<<".format(orig_node_code, astor.to_source(new_node).strip()))
             elif isinstance(new_node, list):
                 log.debug("Converted >>> {} <<< to [[[ {} ]]]".format(orig_node_code, ", ".join(
                     astor.to_source(n).strip() for n in new_node)))
@@ -155,12 +158,14 @@ class TrackedContextTransformer(DebugTransformerMixin, ast.NodeTransformer):
         return node
 
     def __setitem__(self, key, value):
-        log.debug("Context setting {} to {}".format(key, "?UNKNOWN?" if value is None else astor.to_source(value).strip()))
+        log.debug(
+            "Context setting {} to {}".format(key, "?UNKNOWN?" if value is None else astor.to_source(value).strip()))
         self.ctxt[key] = value
 
     def __getitem__(self, item):
         res = self.ctxt[item]
-        log.debug("Context resolving {} to {}".format(item, "?UNKNOWN?" if res is None else astor.to_source(res).strip()))
+        log.debug(
+            "Context resolving {} to {}".format(item, "?UNKNOWN?" if res is None else astor.to_source(res).strip()))
         return res
 
     def __delitem__(self, key):
@@ -189,7 +194,9 @@ class TrackedContextTransformer(DebugTransformerMixin, ast.NodeTransformer):
                             else:  # (_a_, b) = _1_, 2
                                 yield from self._assign(subname, next(iterable_val))
                     except StopIteration:
-                        raise IndexError("Failed to unpack {} into {}, either had too few elements or values after a starred variable".format(val, name))
+                        raise IndexError(
+                            "Failed to unpack {} into {}, either had too few elements or values after a starred variable".format(
+                                val, name))
                 else:
                     yield from self._assign(name, None)
 
@@ -226,9 +233,10 @@ class TrackedContextTransformer(DebugTransformerMixin, ast.NodeTransformer):
                         try:
                             cur_iterable[literal_idx] = val
                         except (IndexError, KeyError):
-                            raise IndexError("Cannot assigned {arr}[{idx}] = {set} where {arr}={arr_val} and {idx}={idx_val}".format(
-                                arr=arr.id, idx=idx, set=val, arr_val=cur_iterable, idx_val=literal_idx
-                            ))
+                            raise IndexError(
+                                "Cannot assigned {arr}[{idx}] = {set} where {arr}={arr_val} and {idx}={idx_val}".format(
+                                    arr=arr.id, idx=idx, set=val, arr_val=cur_iterable, idx_val=literal_idx
+                                ))
                         self[arr.id] = make_ast_from_literal(cur_iterable)
                     else:  # a[???] = val
                         self[arr.id] = None
@@ -284,7 +292,7 @@ class TrackedContextTransformer(DebugTransformerMixin, ast.NodeTransformer):
 
     def visit_FunctionDef(self, node):
         if not self.in_main_func:
-            self.ctxt.push({}, False)
+            self.ctxt.push({})
             node.body = self.nested_visit(node.body, set_conditional_exec=False)
             self.ctxt.pop()
             return self.generic_visit_less(node, 'body')
@@ -350,7 +358,7 @@ class TrackedContextTransformer(DebugTransformerMixin, ast.NodeTransformer):
 def make_function_transformer(transformer_type, name, description, **transformer_kwargs):
     @optional_argument_decorator
     @magic_contract
-    def transform(return_source=False, save_source=True, function_globals=None, **kwargs):
+    def transform(return_source=False, save_source=True, function_globals=None, collapse_iterables=False, explicit_only=False, **kwargs):
         """
         :param return_source: Returns the transformed function's source code instead of compiling it
         :type return_source: bool
@@ -358,6 +366,10 @@ def make_function_transformer(transformer_type, name, description, **transformer
         :type save_source: bool
         :param function_globals: Overridden global name assignments to use when processing the function
         :type function_globals: dict|None
+        :param collapse_iterables: Collapse iterable types
+        :type collapse_iterables: bool
+        :param explicit_only: Whether to use global variables or just keyword and function_globals in the replacement context
+        :type explicit_only: bool
         :param kwargs: Any other environmental variables to provide during unrolling
         :type kwargs: dict
         :return: The transformed function, or its source code if requested
@@ -367,16 +379,23 @@ def make_function_transformer(transformer_type, name, description, **transformer
         @magic_contract(f='Callable', returns='Callable|str')
         def inner(f):
             f_mod, f_body, f_file = function_ast(f)
-            # Grab function globals
-            glbls = f.__globals__
-            # Grab function closure variables
-            if isinstance(f.__closure__, tuple):
-                glbls.update({k: v.cell_contents for k, v in zip(f.__code__.co_freevars, f.__closure__)})
+            if not explicit_only:
+                # Grab function globals
+                glbls = f.__globals__.copy()
+                # Grab function closure variables
+                if isinstance(f.__closure__, tuple):
+                    glbls.update({k: v.cell_contents for k, v in zip(f.__code__.co_freevars, f.__closure__)})
+            else:
+                # Initialize empty context
+                if function_globals is None and len(kwargs) == 0:
+                    log.warning("No global context nor function context. No collapse will occur")
+                glbls = dict()
             # Apply manual globals override
             if function_globals is not None:
                 glbls.update(function_globals)
             # print({k: v for k, v in glbls.items() if k not in globals()})
             trans = transformer_type(DictStack(glbls, kwargs), **transformer_kwargs)
+            trans.collapse_iterables = collapse_iterables
             f_mod.body[0].decorator_list = []
             f_mod = trans.visit(f_mod)
             # print(astor.dump_tree(f_mod))
